pub mod brc_tree {
    #![allow(dead_code)]
    use std::{collections::VecDeque, io::Write, path::Path, vec};

    use indexmap::IndexMap;

    #[derive(Debug, Clone)]
    pub struct Node {
        min: isize,
        max: isize,
        contents: Vec<String>,
        length: isize,
        left: Subtree,
        right: Subtree,
    }

    #[derive(Debug, Clone)]
    pub struct Subtree(Option<Box<Node>>);

    pub fn from_sorted_list(hm: IndexMap<isize, Vec<String>>) -> Subtree {
        let len = hm.len();
        if len == 0 {
            return Subtree(None);
        } else if len == 1 {
            let contents = hm.get_index(0).unwrap().1.to_vec();
            let length = contents.len() as isize;
            return Subtree(Some(Box::new(Node {
                min: *hm.get_index(0).unwrap().0,
                max: *hm.get_index(0).unwrap().0,
                contents,
                length,
                left: Subtree(None),
                right: Subtree(None),
            })));
        }
        let mid = len / 2;
        let left = from_sorted_list(hm.clone().drain(..mid).collect());
        let right = from_sorted_list(hm.clone().drain(mid..).collect());
        let min = match &left.0 {
            Some(node) => node.min,
            None => match &right.0 {
                Some(node) => node.min,
                None => *hm.get_index(0).unwrap().0,
            },
        };
        let max = match &right.0 {
            Some(node) => node.max,
            None => match &left.0 {
                Some(node) => node.max,
                None => *hm.get_index(len - 1).unwrap().0,
            },
        };
        let length = match &left {
            Subtree(Some(node)) => node.length,
            Subtree(None) => 0,
        } + match &right {
            Subtree(Some(node)) => node.length,
            Subtree(None) => 0,
        };
        Subtree(Some(Box::new(Node {
            min,
            max,
            length,
            contents: vec![],
            left,
            right,
        })))
    }
    #[test]
    fn test_fsl() {
        let test_data = indexmap::indexmap! {
            1 => vec!["a".to_string()],
            2 => vec!["b".to_string()],
            3 => vec!["c".to_string()],
            4 => vec!["d".to_string()],
        };
        let tree = from_sorted_list(test_data);
        println!("{:?}", tree);
        println!("{:?}", tree.to_hm());
    }

    impl Subtree {
        pub fn get_all_child_values(&self) -> Vec<String> {
            match &self.0 {
                Some(node) => {
                    let mut left = node.left.get_all_child_values();
                    let right = node.right.get_all_child_values();
                    left.extend(right);
                    left.extend(node.contents.clone());
                    left
                }
                None => vec![],
            }
        }
        pub fn to_hm(&self) -> IndexMap<(isize, isize), Vec<String>> {
            let mut hm: IndexMap<(isize, isize), Vec<String>> = IndexMap::new();
            if let Some(node) = &self.0 {
                let left = node.left.to_hm();
                let right = node.right.to_hm();
                for (key, value) in left.iter() {
                    hm.insert(*key, value.clone());
                }
                for (key, value) in right.iter() {
                    hm.insert(*key, value.clone());
                }
                let value = self.get_all_child_values();
                hm.insert(self.range(), value);
            }
            hm
        }

        pub fn write_structure(&self, mut file: &std::fs::File) -> anyhow::Result<()> {
            if let Some(node) = &self.0 {
                file.write_all(format!("{},{}[{}](", node.min, node.max, node.length).as_bytes())?;
                node.left.write_structure(file)?;
                file.write_all(")(".as_bytes())?;
                node.right.write_structure(file)?;
                file.write_all(");".as_bytes())?;
            }
            Ok(())
        }

        pub fn covers(&self, min: isize, max: isize) -> Vec<Subtree> {
            // Returns the set of nodes that covers the entire range
            match &self.0 {
                Some(node) => {
                    if node.min > max || node.max < min {
                        return vec![];
                    }
                    if min <= node.min && max >= node.max {
                        return vec![self.clone()];
                    }
                    let mut left = node.left.covers(min, max);
                    let right = node.right.covers(min, max);
                    left.extend(right);
                    left
                }
                None => vec![],
            }
        }

        pub fn range(&self) -> (isize, isize) {
            match &self.0 {
                Some(node) => (node.min, node.max),
                None => (0, 0),
            }
        }

        pub fn range_level(&self) -> (isize, isize, isize) {
            match &self.0 {
                Some(node) => (node.min, node.max, self.level()),
                None => (0, 0, 0),
            }
        }

        pub fn level(&self) -> isize {
            match &self.0 {
                Some(node) => {
                    if node.max == node.min {
                        return 0;
                    }
                    let left = node.left.level();
                    let right = node.right.level();
                    1 + left.max(right)
                }
                None => -1,
            }
        }
    }
    #[test]
    fn test_possible_covers() {
        let tree = parse_from_file(std::path::Path::new("../brc_table.partitions")).unwrap();
        let mut covers = tree.covers(0, 111000);
        println!("{:?}", tree);
        println!("{:?}", covers.iter().map(|x| x.range()).collect::<Vec<_>>());
        covers.sort_by_key(|a| a.level());
        println!(
            "{:?}",
            brc_to_urc(covers)
                .iter()
                .map(|x| x.range())
                .collect::<Vec<_>>()
        );
    }
    pub fn parse_from_file(file_name: &Path) -> anyhow::Result<Subtree> {
        let contents = std::fs::read_to_string(file_name)?;
        parse_from_string(contents)
    }

    impl PartialEq for Subtree {
        fn eq(&self, other: &Self) -> bool {
            self.range() == other.range()
        }
    }

    pub fn brc_to_urc(vec: Vec<Subtree>) -> Vec<Subtree> {
        // assume vec is sorted by level
        let mut deq = VecDeque::from(vec);
        let max_level = deq[deq.len() - 1].level();
        let mut urc: Vec<Subtree> = vec![];
        let mut seen = VecDeque::from_iter(0..=max_level + 1);
        while !deq.is_empty() {
            let node = deq.pop_front().unwrap().to_owned();
            if node.0.is_none() {
                continue;
            }
            match node.level() {
                i if i < seen[0] => {
                    if !urc.contains(&node) {
                        urc.push(node);
                    }
                }
                i if i == seen[0] => {
                    seen.pop_front();
                    if !urc.contains(&node) {
                        urc.push(node);
                    }
                }
                _ => {
                    deq.push_front(node.clone().0.unwrap().left);
                    deq.push_front(node.0.unwrap().right);
                }
            }
        }
        urc
    }

    pub fn parse_from_string(s: String) -> anyhow::Result<Subtree> {
        let s = s.trim();
        if s.is_empty() {
            return Ok(Subtree(None));
        }
        let (min_max, rest_before_length) = s.split_at(s.find('[').unwrap());
        let (min, max) = {
            let mut parts = min_max.split(',');
            (
                parts.next().unwrap().parse::<isize>()?,
                parts.next().unwrap().parse::<isize>()?,
            )
        };
        let (length_str, rest) = rest_before_length.split_at(rest_before_length.find(']').unwrap());
        let length = length_str[1..].parse::<isize>()?;

        // Find the closing parenthesis for the left subtree
        let mut balance = 1; // Start at 1 because we're already inside the first parenthesis
        let mut left_end = 0;
        for (i, c) in rest[2..].chars().enumerate() {
            match c {
                '(' => balance += 1,
                ')' => balance -= 1,
                _ => {}
            }
            if balance == 0 {
                left_end = i + 2; // +2 to account for the substring offset
                break;
            }
        }

        // The right subtree starts after the )( separator
        let right_start = left_end + 2; // +2 to skip the ")("

        let left = parse_from_string(rest[2..left_end].to_string())?;
        let right = parse_from_string(rest[right_start..rest.len() - 2].to_string())?;

        Ok(Subtree(Some(Box::new(Node {
            min,
            max,
            length,
            contents: vec![],
            left,
            right,
        }))))
    }

    #[test]
    fn test_brc_to_urc() {
        let tree = from_sorted_list(indexmap::indexmap! {
            0 => vec!["a".to_string()],
            1 => vec!["a".to_string()],
            2 => vec!["b".to_string()],
            3 => vec!["c".to_string()],
            4 => vec!["d".to_string()],
            5 => vec!["e".to_string()],
            6 => vec!["f".to_string()],
            7 => vec!["g".to_string()],
        });
        let urc = brc_to_urc(tree.covers(2, 7));
        println!("{:?}", urc.iter().map(|x| x.range()).collect::<Vec<_>>());
    }

    impl TryInto<Subtree> for String {
        type Error = anyhow::Error;
        fn try_into(self) -> Result<Subtree, Self::Error> {
            parse_from_string(self)
        }
    }

    #[test]
    fn test_parse() {
        let tree = parse_from_file("../brc_table.partitions".as_ref()).unwrap();
        println!("{:?}", tree);
    }
}
