use std::collections::HashMap;

use anyhow::{anyhow, Result};
use indexmap::IndexMap;
use log::{debug, info, warn};

use crypto_utils::chacha::encrypt;
use crypto_utils::hash::{hash_with_counter, prf};
use crypto_utils::keys::load_from_file;

use crate::brc::brc_tree;

pub fn init(
    filename: String,
    key_option: Option<String>,
    index_option: Option<usize>,
    database_address: String,
) -> Result<()> {
    if key_option.is_none() && index_option.is_none() {
        info!("Key not provided, will use the first avaliable")
    }
    if key_option.is_some() && index_option.is_some() {
        warn!("Both key and index provided!");
        return Err(anyhow!("Both key and index provided"));
    }
    debug!("Given file {}, searching whether it exists", filename);
    let filepath = std::path::Path::new(&filename);
    // verify that file exists
    if !&filepath.exists() {
        warn!("{:?} cannot be found", filepath);
        return Err(anyhow!("Provided file cannot be found {:?}", filepath));
    }
    debug!("File exists");
    let mut hm: IndexMap<isize, Vec<String>> = IndexMap::new();
    // let mut hm : HashMap<isize, Vec<String>> = HashMap::new();
    let mut reader = csv::Reader::from_path(filepath)?;

    let enc_key = load_from_file("enc.key")?;
    let prf_key = load_from_file("prf.key")?;
    let index = match key_option {
        Some(s) => {
            let headers = reader.headers()?;
            let mut ret: usize = usize::MAX;
            for (index, word) in headers.iter().enumerate() {
                if word.eq_ignore_ascii_case(&s) {
                    ret = index;
                    break;
                }
            }
            if ret == usize::MAX {
                warn!("{} cannot be found in headers", s);
                return Err(anyhow!("{} not found in headers", s));
            }
            ret
        }
        None => match index_option {
            Some(i) => {
                let headers = reader.headers()?;
                if i > headers.len() {
                    warn!(
                        "Header has length {}, which is less than {}",
                        headers.len(),
                        i
                    );
                    return Err(anyhow!("Index out of bounds"));
                }
                i
            }
            None => 0,
        },
    };
    for record in reader.records() {
        let record = record?;
        let key = record.get(index).unwrap().parse();
        if key.is_err() {
            warn!("Could not parse key {}", record.get(index).unwrap());
            return Err(anyhow!("Could not parse key"));
        }
        hm.entry(key.unwrap())
            .or_default()
            .push(record.iter().collect::<Vec<&str>>().join(","));
    }
    debug!("{:?}", hm);
    hm.sort_keys();
    let hashtable = brc_tree::from_sorted_list(hm);
    let mut json: HashMap<String, String> = HashMap::new();
    let file = std::fs::File::create("brc_table.partitions")?;
    hashtable.write_structure(&file)?;
    for (key, value) in hashtable.to_hm().iter() {
        for (counter, record) in value.iter().enumerate() {
            let encrypted = encrypt(&enc_key, record.to_string())?;
            let prfd_key = prf(&prf_key, format!("{:?}", key))?;
            let hashed = hash_with_counter(prfd_key, counter)?;
            json.insert(hashed, encrypted);
        }
    }
    reqwest::blocking::Client::new()
        .post(database_address)
        .json(&json)
        .send()?;
    for i in json.iter() {
        println!("{:?}", i);
    }
    println!("{}", json.len());
    Ok(())
}
