use std::collections::HashMap;

use crate::pools::pool_get_metadata;
use crate::pools::search_availability;
use crate::pools::MetaData;
use anyhow::{anyhow, Result};

use crypto_utils::chacha::decrypt_with_action_and_validity;
use crypto_utils::chacha::ChaCha20Rng;
// This is just a current implementation of the outputs
use log::info;
use url::Url;

use crypto_utils::chacha::decrypt;
use crypto_utils::chacha::rand;

use crypto_utils::hash::prf;
use crypto_utils::keys::load_from_file;
use rand::seq::SliceRandom;

use crate::brc::brc_tree;
use crate::brc::brc_tree::Subtree;

/// Search function
pub async fn search(
    min: isize,
    max: isize,
    address: Vec<Url>,
    urf: bool,
    output_path: String,
    rng: &mut ChaCha20Rng,
) -> Result<()> {
    // info!("Searching for range between {} and {}", min, max);
    let enc_key = load_from_file("enc.key")?;
    let prf_key = load_from_file("prf.key")?;
    // let burl = Url::parse(&address)?;
    let metadata: MetaData = get_metadata(address.clone(), &enc_key, rng).await?;
    let btrees = get_btrees(address, &enc_key).await?;
    let mut existing_covers: HashMap<Url, Vec<(isize, isize)>> = HashMap::new();
    for (url, v_subtrees) in btrees {
        for subtree in v_subtrees {
            let mut covers = subtree.covers(min, max);
            if !covers.is_empty() {
                if urf {
                    covers.sort_by_key(|a| a.level());
                    covers = brc_tree::brc_to_urc(covers);
                }
                existing_covers
                    .entry(url.clone())
                    .or_default()
                    .extend(covers.iter().map(|x| x.to_owned().range()));
            }
        }
    }
    for covers in existing_covers.values_mut() {
        covers.sort();
        covers.dedup();
    }
    // println!("{:?}", existing_covers.values());
    let mut queries: Vec<(Url, String)> = Vec::new();
    for (urls, covers) in existing_covers {
        for cover in covers {
            println!("{}: {:?}", urls, cover);
            queries.push((urls.clone(), prf(&prf_key, format!("{:?}", cover))?));
        }
    }
    queries.shuffle(rng);
    let mut futures = Vec::new();
    // different servers get different keys
    for (urls, key) in queries {
        let mut search_url = urls.clone();
        search_url.set_path(&format!("key/{}", key.clone()));
        futures.push(tokio::spawn(async move {
            let client = reqwest::Client::new();
            // info!("Searching for {}", key);
            let resp = client
                .get(search_url)
                .send()
                .await?
                .json::<Vec<String>>()
                .await?;
            Ok::<Vec<String>, reqwest::Error>(resp)
        }));
    }
    let joined_futures = futures::future::join_all(futures).await;

    let mut fnl: Vec<(isize, String, bool, u64)> = Vec::new();
    for ciphertext_v in joined_futures {
        for ciphertext in ciphertext_v?? {
            let decrypted = decrypt_with_action_and_validity(&enc_key, ciphertext)?;
            if decrypted.2 != 0 {
                let split: Vec<String> = decrypted
                    .0
                    .split(",")
                    .map(|x| x.to_string())
                    .collect::<Vec<String>>();
                let parsed = split
                    .get(metadata.index as usize)
                    .ok_or(anyhow!("Index out of range"))?
                    .parse::<isize>()?;
                fnl.push((parsed, decrypted.0, decrypted.1, decrypted.2));
            } else {
                info!("Ignoring {:?}", decrypted);
            }
        }
    }
    fnl.sort_by(|a, b| a.3.cmp(&b.3));
    fnl.dedup();
    // remove elements that are at some point removed
    let mut hm: HashMap<isize, Vec<String>> = HashMap::new();
    for (key, value, add, _) in fnl {
        if add {
            hm.entry(key).or_default().push(value);
        } else {
            info!("Removing {:?}", key);
            hm.remove(&key);
        }
    }
    for value in hm.values_mut() {
        value.sort();
        value.dedup();
    }
    // write to csv file
    let write_path = std::fs::OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(output_path)?;
    let mut wtr = csv::Writer::from_writer(write_path);
    wtr.write_record(&metadata.header)?;
    let mut hm_keys: Vec<isize> = hm.keys().clone().map(|x| x.to_owned()).collect();
    hm_keys.sort();
    for key in hm_keys {
        for value in hm.get(&key).unwrap() {
            wtr.write_record(value.split(","))?;
        }
    }
    Ok(())
}

async fn get_btrees(urls: Vec<Url>, key: &[u8]) -> Result<HashMap<Url, Vec<Subtree>>> {
    let btree_map = search_availability(urls).await?;
    let mut return_map = HashMap::new();
    for (url, vec) in btree_map {
        let mut btrees = Vec::new();
        for value in vec.iter() {
            let decrypted = decrypt(key, value.to_string())?;
            let subtree = brc_tree::parse_from_string(decrypted)?;
            btrees.push(subtree);
        }
        return_map.insert(url, btrees);
    }
    Ok(return_map)
}

pub async fn get_metadata(
    base_url: Vec<Url>,
    key: &[u8],
    rng: &mut ChaCha20Rng,
) -> Result<MetaData> {
    let encrypted = pool_get_metadata(base_url, rng).await?;
    let decrypted = decrypt(key, encrypted)?;
    Ok(serde_json::from_str(&decrypted)?)
}
