use std::collections::HashMap;

use anyhow::{anyhow, Result};
use crypto_utils::chacha::rand::{Rng, SeedableRng};
use indexmap::IndexMap;
use log::{debug, info, warn};

use crypto_utils::chacha::{
    encrypt, encrypt_with_action_and_validity, generate_nonce_string, ChaCha20Rng,
};
use crypto_utils::hash::{hash_with_counter_and_nonce, prf};
use crypto_utils::keys::load_from_file;
use serde::{Deserialize, Serialize};

use crate::brc::brc_tree;

pub fn generate_one(
    hm: &mut IndexMap<isize, Vec<(String, bool)>>,
    rng: &mut ChaCha20Rng,
) -> Result<()> {
    let mut min: i64 = ((hm.keys().min().unwrap_or(&0).to_owned()) * 95 / 100).try_into()?;
    let mut max: i64 = (hm.keys().max().unwrap_or(&100).to_owned() * 105 / 100).try_into()?;
    let row_length = hm
        .values()
        .next()
        .unwrap_or(&vec![("a".repeat(100), false)])[0]
        .0
        .len();
    let min_row_length = (row_length * 3) / 10;
    let max_row_length = (row_length * 20) / 10;
    let mut key;
    loop {
        info!("{},{}", min, max);
        key = rng.random_range(min..max) as isize;
        if !hm.contains_key(&key) {
            break;
        } else {
            min -= 1;
            max += 1;
        }
    }
    let mut vec: Vec<(String, bool)> = Vec::new();
    let repeat = "a".repeat(rng.random_range(min_row_length..max_row_length));
    vec.push((repeat, false));
    hm.insert(key, vec);
    Ok(())
}

#[derive(Serialize, Deserialize)]
struct Init {
    level: usize,
    db: HashMap<String, String>,
    metadata: String,
    nonce: String,
    btree: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct MetaData {
    pub index: i32,
    pub header: Vec<String>,
    pub row_length: i32,
}

pub fn init(
    filename: String,
    key_option: Option<String>,
    index_option: Option<usize>,
    database_address: String,
) -> Result<()> {
    if key_option.is_none() && index_option.is_none() {
        info!("Key not provided, will use the first avaliable")
    }
    if key_option.is_some() && index_option.is_some() {
        warn!("Both key and index provided!");
        return Err(anyhow!("Both key and index provided"));
    }
    debug!("Given file {}, searching whether it exists", filename);
    let filepath = std::path::Path::new(&filename);
    // verify that file exists
    if !&filepath.exists() {
        warn!("{:?} cannot be found", filepath);
        return Err(anyhow!("Provided file cannot be found {:?}", filepath));
    }
    debug!("File exists");
    let mut hm: IndexMap<isize, Vec<(String, bool)>> = IndexMap::new();
    let mut reader = csv::Reader::from_path(filepath)?;
    let enc_key = load_from_file("enc.key")?;
    let prf_key = load_from_file("prf.key")?;
    let index = match key_option {
        Some(s) => {
            let headers = reader.headers()?;
            let mut ret: usize = usize::MAX;
            for (index, word) in headers.iter().enumerate() {
                if word.eq_ignore_ascii_case(&s) {
                    ret = index;
                    break;
                }
            }
            if ret == usize::MAX {
                warn!("{} cannot be found in headers", s);
                return Err(anyhow!("{} not found in headers", s));
            }
            ret
        }
        None => match index_option {
            Some(i) => {
                let headers = reader.headers()?;
                if i > headers.len() {
                    warn!(
                        "Header has length {}, which is less than {}",
                        headers.len(),
                        i
                    );
                    return Err(anyhow!("Index out of bounds"));
                }
                i
            }
            None => 0,
        },
    };
    let meta: MetaData = MetaData {
        index: index as i32,
        header: reader.headers()?.iter().map(|x| x.to_string()).collect(),
        row_length: reader.headers()?.len() as i32,
    };
    let meta_string = encrypt(&enc_key, serde_json::to_string(&meta)?)?;
    for record in reader.records() {
        let record = record?;
        let key = record.get(index).unwrap().parse();
        let record_vec = record.iter().collect::<Vec<&str>>().join(",");

        if key.is_err() {
            warn!("Could not parse key {}", record.get(index).unwrap());
            return Err(anyhow!("Could not parse key"));
        }
        hm.entry(key.unwrap()).or_default().push((record_vec, true));
    }
    // populate hashmap until it is a power of 2
    let level = (hm.len() as f64).log2().ceil() as usize;
    let diff = (2_isize.pow(level as u32) as usize - hm.len()) as isize;
    debug!(
        "{} distinct keys seen, we will insert {} keys",
        hm.len(),
        diff
    );
    let mut rng = ChaCha20Rng::from_os_rng();
    for _ in 0..diff {
        generate_one(&mut hm, &mut rng)?;
    }
    hm.sort_keys();
    debug!("{:?}", hm);
    info!("{:?}", hm.len());
    let subtree = brc_tree::from_sorted_list(hm);
    let nonce = generate_nonce_string();
    let mut json: HashMap<String, String> = HashMap::new();
    for (key, value) in subtree.to_hm().iter() {
        let prfd_key = prf(&prf_key, format!("{:?}", key))?;
        for (counter, record) in value.iter().enumerate() {
            let (pt, va) = record;
            let encrypted =
                encrypt_with_action_and_validity(&enc_key, pt.to_owned(), true, va.to_owned())?;
            let hashed = hash_with_counter_and_nonce(prfd_key.clone(), counter, &nonce)?;
            json.insert(hashed, encrypted);
        }
    }
    let init_json = Init {
        level,
        db: json,
        nonce,
        metadata: meta_string,
        btree: encrypt(&enc_key, subtree.write_string())?,
    };
    let client = reqwest::blocking::Client::builder()
        .timeout(std::time::Duration::from_secs(60 * 30))
        .build()?;
    client.post(database_address).json(&init_json).send()?;
    Ok(())
}
