use std::collections::HashMap;
use std::iter;

use crate::brc::brc_tree;
use crate::initialize::{generate_one, MetaData};
use crate::search::get_metadata;
use anyhow::{anyhow, Result};
use crypto_utils::chacha::rand::SeedableRng;
use crypto_utils::chacha::{
    decrypt_with_action_and_validity, encrypt, encrypt_with_action_and_time,
    encrypt_with_action_and_validity, generate_nonce_string, ChaCha20Rng,
};
use crypto_utils::hash::{hash_with_counter_and_nonce, prf};
use crypto_utils::keys::load_from_file;
use indexmap::{indexmap, IndexMap};
use itertools::Itertools;
use log::{debug, info};
use serde::{Deserialize, Serialize};
use url::Url;

#[derive(Serialize, Deserialize, Debug)]
pub struct InsertResponse {
    sda: bool,
    values: Vec<String>,
    sda_level: usize,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct InsertRequest {
    nonce: String,
    db: HashMap<String, String>,
    btree: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct SdaRequest {
    level: usize,
    nonce: String,
    db: HashMap<String, String>,
    btree: String,
}
pub fn insert(value: String, address: String) -> Result<()> {
    let base_url: Url = Url::parse(&address)?;
    let enc_key = load_from_file("enc.key")?;
    let prf_key = load_from_file("prf.key")?;
    let metadata = get_metadata(base_url.clone(), &enc_key)?;
    let index_vec = value.split(",").collect::<Vec<&str>>();
    let key = index_vec
        .get(metadata.index as usize)
        .ok_or(anyhow!("Index out of bounds"))?
        .to_owned()
        .to_owned()
        .parse()?;
    let btree = brc_tree::from_sorted_list(indexmap! {key => vec![(value.clone(), true)]});
    info!("Inserting into {} with values {}", key, value);
    let encrypted_btree = encrypt(&enc_key, btree.to_string())?;
    let mut item_map: HashMap<String, String> = HashMap::new();
    let nonce = generate_nonce_string();
    item_map.insert(
        hash_with_counter_and_nonce(prf(&prf_key, format!("{:?}", (key, key)))?, 0, &nonce)?,
        encrypt_with_action_and_validity(&enc_key, value, true, true)?,
    );
    let insert_request_json = InsertRequest {
        nonce,
        db: item_map,
        btree: encrypted_btree,
    };
    let resp = reqwest::blocking::Client::new()
        .put(base_url.join("key")?)
        .json(&insert_request_json)
        .send()
        .unwrap()
        .json::<InsertResponse>()?;
    debug!("Got response {:?}", resp);
    sda(resp, metadata, address)
}

pub fn sda(response: InsertResponse, metadata: MetaData, address: String) -> Result<()> {
    if !response.sda {
        return Ok(());
    }
    info!(
        "Starting SDA Level {} with {} elements",
        response.sda_level,
        response.values.len()
    );
    let enc_key = load_from_file("enc.key")?;
    let prf_key = load_from_file("prf.key")?;
    let value_parsed: Result<Vec<(String, bool, u64)>, _> = response
        .values
        .iter()
        .map(|x| decrypt_with_action_and_validity(&enc_key, x.to_owned()))
        .collect();
    info!("Values parsed");
    let mut values = value_parsed?
        .iter()
        .filter(|x| x.2 != 0)
        .map(|x| x.to_owned())
        .unique()
        .collect::<Vec<_>>();
    info!("Values Length: {}", values.len());
    values.sort_by_key(|x| x.2);
    info!("Sorted");
    let mut hm: IndexMap<isize, Vec<(String, bool)>> = IndexMap::new();
    let mut actions: HashMap<String, (bool, u64)> = HashMap::new();
    for (value, add, timer) in values {
        // filter out nonvalid items
        if timer == 0 {
            continue;
        }
        let split: Vec<String> = value.split(",").map(|x| x.to_string()).collect();
        let parsed = split
            .get(metadata.index as usize)
            .ok_or(anyhow!("Index out of range"))?
            .parse::<isize>()?;
        hm.entry(parsed).or_default().push((value.clone(), true));
        if !add {
            info!("add is false with value {}", value);
        }
        actions.insert(value, (add, timer));
    }
    info!(
        "Total elements: {}",
        hm.values().map(|x| x.len()).sum::<usize>()
    );
    let target_amount = 2_isize.pow(response.sda_level as u32) as usize;
    let diff = target_amount - hm.len();
    info!(
        "Generating {} dummy records (target: {}, curr: {})",
        diff,
        target_amount,
        hm.len()
    );
    let mut rng = ChaCha20Rng::from_os_rng();
    for i in 0..diff {
        generate_one(&mut hm, &mut rng)?;
        info!("{}", i);
    }
    hm.sort_keys();
    let subtree = brc_tree::from_sorted_list(hm);
    let mut json: HashMap<String, String> = HashMap::new();
    let mut real = 0;
    let mut fake = 0;
    let nonce = generate_nonce_string();
    for (key, value) in subtree.to_hm().iter_mut() {
        for (counter, record) in value.iter().enumerate() {
            let (pt, va) = record;

            let encrypted = match *va {
                true => {
                    real += 1;
                    let (action, timer) = actions
                        .get(pt)
                        .ok_or(anyhow!("Action should have been registered prior"))?;
                    encrypt_with_action_and_time(&enc_key, pt.to_owned(), *action, *timer)?
                }
                false => {
                    fake += 1;
                    encrypt_with_action_and_validity(&enc_key, pt.to_owned(), true, false)?
                }
            };
            let prfd_key = prf(&prf_key, format!("{:?}", key))?;
            let hashed = hash_with_counter_and_nonce(prfd_key, counter, &nonce)?;
            json.insert(hashed, encrypted);
        }
    }
    info!("Real {} vs Fake {}", real, fake);
    let insert = SdaRequest {
        level: response.sda_level,
        nonce,
        db: json,
        btree: encrypt(&enc_key, subtree.to_string())?,
    };
    let client = reqwest::blocking::Client::builder()
        .timeout(std::time::Duration::from_secs(30))
        .build()?;
    let resp = client
        .put(Url::parse(&address)?.join("sda")?)
        .json(&insert)
        .send()
        .unwrap()
        .json::<InsertResponse>()?;
    info!("Writing to sda level {}", response.sda_level);
    if resp.sda {
        sda(resp, metadata, address)
    } else {
        Ok(())
    }
}

pub fn delete(key: String, address: String) -> Result<()> {
    let base_url: Url = Url::parse(&address)?;
    let enc_key = load_from_file("enc.key")?;
    let prf_key = load_from_file("prf.key")?;
    let metadata = get_metadata(base_url.clone(), &enc_key)?;
    let key_int: isize = key.parse()?;
    let prfd_key = prf(&prf_key, format!("{:?}", (key_int, key_int)))?;
    let mut plaintext_iter = iter::repeat("a")
        .take(metadata.row_length as usize)
        .collect::<Vec<&str>>();
    plaintext_iter[metadata.index as usize] = &key;
    let plaintext = plaintext_iter.join(",");
    let stree = brc_tree::from_sorted_list(indexmap! {
        key.parse::<isize>()? => vec![(plaintext.clone(), true)]
    });
    info!("stree {}", stree);
    let btree = encrypt(&enc_key, stree.to_string())?;
    let mut db: HashMap<String, String> = HashMap::new();
    let nonce = generate_nonce_string();
    db.insert(
        hash_with_counter_and_nonce(prfd_key, 0, &nonce)?,
        encrypt_with_action_and_validity(&enc_key, plaintext, false, true)?,
    );
    let delete_request = InsertRequest { nonce, db, btree };
    info!(
        "Deleting key {}, with request data {:?}",
        key, delete_request
    );
    let resp = reqwest::blocking::Client::new()
        .put(base_url.join("key")?)
        .json(&delete_request)
        .send()
        .unwrap()
        .json::<InsertResponse>()?;
    debug!("Got response {:?}", resp);
    sda(resp, metadata, address)
}
