use std::collections::HashMap;
use std::io::prelude::*;

use crate::initialize::MetaData;
use anyhow::{anyhow, Result};

use crypto_utils::chacha::decrypt_with_action_and_validity;
// This is just a current implementation of the outputs
use crypto_utils::chacha::rand::SeedableRng;
use log::info;
use url::Url;

use crypto_utils::chacha::decrypt;
use crypto_utils::chacha::rand;
use crypto_utils::chacha::ChaCha20Rng;

use crypto_utils::hash::prf;
use crypto_utils::keys::load_from_file;
use rand::seq::SliceRandom;

use crate::brc::brc_tree;
use crate::brc::brc_tree::Subtree;

/// Search function
pub fn search(
    min: isize,
    max: isize,
    address: String,
    urf: bool,
    output_path: String,
) -> Result<()> {
    info!("Searching for range between {} and {}", min, max);
    let enc_key = load_from_file("enc.key")?;
    let prf_key = load_from_file("prf.key")?;
    let mut burl = Url::parse(&address)?;
    let metadata = get_metadata(burl.clone(), &enc_key)?;
    let btrees = get_btrees(burl.clone(), &enc_key)?;
    let mut covers = Vec::new();
    for subtree in btrees {
        // load covers from brc tree
        let mut c = subtree.covers(min, max);
        if c.is_empty() {
            // ignore cases where it is out of bounds
            continue;
        }
        if urf {
            c.sort_by_key(|a| a.level());
            c = brc_tree::brc_to_urc(c);
        }
        covers.extend(c);
    }
    info!(
        "Best range covers with height: {:?}",
        covers.iter().map(|x| x.range_level()).collect::<Vec<_>>()
    );
    let search_keys: Result<Vec<String>> = covers
        .iter()
        .map(|x| prf(&prf_key, format!("{:?}", x.range())))
        .collect();
    let mut keys = search_keys?;
    // shuffle key order
    keys.shuffle(&mut ChaCha20Rng::from_os_rng());
    let mut fnl: Vec<(isize, String, bool, u64)> = Vec::new();
    for key in keys {
        info!("Searching for key {}", key);
        burl.set_path(&format!("key/{}", key));
        let resp = reqwest::blocking::get(burl.clone())?.json::<Vec<String>>()?;
        for ciphertext in resp.iter() {
            let decrypted = decrypt_with_action_and_validity(&enc_key, ciphertext.to_string())?;
            if decrypted.2 != 0 {
                let split: Vec<String> = decrypted
                    .0
                    .split(",")
                    .map(|x| x.to_string())
                    .collect::<Vec<String>>();
                let parsed = split
                    .get(metadata.index as usize)
                    .ok_or(anyhow!("Index out of range"))?
                    .parse::<isize>()?;
                fnl.push((parsed, decrypted.0, decrypted.1, decrypted.2));
            } else {
                info!("Ignoring {:?}", decrypted);
            }
        }
    }
    // sort edits by timeline
    fnl.sort_by(|a, b| a.3.cmp(&b.3));
    fnl.dedup();
    // remove elements that are at some point removed
    let mut hm: HashMap<isize, Vec<String>> = HashMap::new();
    for (key, value, add, _) in fnl {
        if add {
            // info!("Adding {:?} {:?}", key, value);
            hm.entry(key).or_default().push(value);
        } else {
            // info!("Removing {:?}", key);
            hm.remove(&key);
        }
    }
    for value in hm.values_mut() {
        value.sort();
        value.dedup();
    }
    // write to csv file
    let write_path = std::fs::OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(output_path)?;
    let mut wtr = csv::Writer::from_writer(write_path);
    wtr.write_record(&metadata.header)?;
    let mut hm_keys: Vec<isize> = hm.keys().clone().map(|x| x.to_owned()).collect();
    hm_keys.sort();
    for key in hm_keys {
        for value in hm.get(&key).unwrap() {
            wtr.write_record(value.split(","))?;
        }
    }
    Ok(())
}

fn get_btrees(mut base_url: Url, key: &[u8]) -> Result<Vec<Subtree>> {
    base_url.set_path("/btrees");
    let resp = reqwest::blocking::get(base_url)?.json::<Vec<String>>()?;
    let mut btrees = Vec::new();
    for value in resp.iter() {
        let decrypted = decrypt(key, value.to_string())?;
        let subtree = brc_tree::parse_from_string(decrypted)?;
        btrees.push(subtree);
    }
    Ok(btrees)
}

#[test]
fn test_get_btrees() {
    let base_url = Url::parse("http://localhost:8080").unwrap();
    let enc_key = load_from_file("enc.key").unwrap();
    let btrees = get_btrees(base_url, &enc_key).unwrap();
    println!("{:?}", btrees);
}

pub fn get_metadata(mut base_url: Url, key: &[u8]) -> Result<MetaData> {
    base_url.set_path("/metadata");
    let mut buf = String::new();
    let _ = reqwest::blocking::get(base_url)?.read_to_string(&mut buf)?;
    let decrypted = decrypt(key, buf)?;
    Ok(serde_json::from_str(&decrypted)?)
}

#[test]
fn test_get_metadata() {
    let base_url = Url::parse("http://localhost:8080").unwrap();
    let enc_key = load_from_file("enc.key").unwrap();
    let meta = get_metadata(base_url, &enc_key).unwrap();
    println!("{:?}", meta);
}
