use std::collections::HashMap;

use anyhow::{anyhow, Result};
use log::{debug, info, warn};

use crypto_utils::chacha::encrypt;
use crypto_utils::hash::{hash_with_counter, prf};
use crypto_utils::keys::load_from_file;

pub fn init(
    filename: String,
    key_option: Option<String>,
    index_option: Option<usize>,
    database_address: String,
) -> Result<()> {
    if key_option.is_none() && index_option.is_none() {
        info!("Key not provided, will use the first avaliable")
    }
    if key_option.is_some() && index_option.is_some() {
        warn!("Both key and index provided!");
        return Err(anyhow!("Both key and index provided"));
    }
    debug!("Given file {}, searching whether it exists", filename);
    let filepath = std::path::Path::new(&filename);
    // verify that file exists
    if !&filepath.exists() {
        warn!("{:?} cannot be found", filepath);
        return Err(anyhow!("Provided file cannot be found {:?}", filepath));
    }
    debug!("File exists");

    // we only support csv files for now
    let mut hm: HashMap<String, String> = HashMap::new();
    let mut seen: HashMap<String, usize> = HashMap::new();
    let mut reader = csv::Reader::from_path(filepath)?;

    let enc_key = load_from_file("enc.key")?;
    let prf_key = load_from_file("prf.key")?;
    let index = match key_option {
        Some(s) => {
            let headers = reader.headers()?;
            let mut ret: usize = usize::MAX;
            for (index, word) in headers.iter().enumerate() {
                if word.eq_ignore_ascii_case(&s) {
                    ret = index;
                    break;
                }
            }
            if ret == usize::MAX {
                warn!("{} cannot be found in headers", s);
                return Err(anyhow!("{} not found in headers", s));
            }
            ret
        }
        None => match index_option {
            Some(i) => {
                let headers = reader.headers()?;
                if i > headers.len() {
                    warn!(
                        "Header has length {}, which is less than {}",
                        headers.len(),
                        i
                    );
                    return Err(anyhow!(
                        "Header has length {}, less than the provided key {}",
                        headers.len(),
                        i
                    ));
                }
                i
            }
            None => {
                info!("Defaulting to using 0 index as search index");
                0
            }
        },
    };
    info!("Started hashing runs");
    for result in reader.records() {
        let record = result?;
        let prfd_key = prf(
            &prf_key,
            record
                .get(index)
                .ok_or(anyhow!("Failed to get element"))?
                .to_string(),
        )?;
        let seen_amnt = seen.entry(prfd_key.clone()).or_insert(0);
        hm.insert(
            hash_with_counter(prfd_key, *seen_amnt)?,
            encrypt(&enc_key, record.iter().collect::<Vec<&str>>().join(","))?,
        );
        *seen_amnt += 1;
    }
    info!("Finished hashing runs");
    reqwest::blocking::Client::new()
        .post(database_address)
        .json(&hm)
        .send()?;
    Ok(())
}
