pub mod chacha {
    use core::str;
    use std::time::{Duration, SystemTime};

    use anyhow::{anyhow, Result};
    use chacha20poly1305::{
        aead::{Aead, OsRng},
        AeadCore, KeyInit, XChaCha20Poly1305, XNonce,
    };

    use once_cell::sync::Lazy;
    pub use rand;

    pub use rand_chacha::ChaCha20Rng;

    pub use base64::prelude::*;
    use regex::Regex;

    pub fn generate_nonce_string() -> String {
        BASE64_STANDARD.encode(XChaCha20Poly1305::generate_nonce(&mut OsRng))
    }

    pub fn encrypt(key: &[u8], plaintext: String) -> Result<String> {
        let cipher = Lazy::new(|| XChaCha20Poly1305::new_from_slice(key).unwrap());
        let nonce = XChaCha20Poly1305::generate_nonce(&mut OsRng);
        let ciphertext = match cipher.encrypt(&nonce, plaintext.as_bytes()) {
            Ok(v) => Ok(v),
            Err(e) => Err(anyhow!("failed to encrypt because {}", e.to_string())),
        }?;
        Ok(BASE64_STANDARD.encode(nonce) + &BASE64_STANDARD.encode(ciphertext))
    }

    pub fn encrypt_with_action_and_validity(
        key: &[u8],
        plaintext: String,
        action: bool,
        validity: bool,
    ) -> Result<String> {
        let time_as_seconds = match validity {
            true => SystemTime::now().duration_since(SystemTime::UNIX_EPOCH),
            false => Ok(Duration::from_secs(0)),
        }?
        .as_secs();
        let encrypt_string = format!("({}:{}:{})", plaintext, action as i64, time_as_seconds);
        encrypt(key, encrypt_string)
    }

    pub fn encrypt_with_action_and_time(
        key: &[u8],
        plaintext: String,
        action: bool,
        time: u64,
    ) -> Result<String> {
        let encrypt_string = format!("({}:{}:{})", plaintext, action as i64, time);
        encrypt(key, encrypt_string)
    }

    pub fn decrypt_with_action_and_validity(
        key: &[u8],
        ciphertext: String,
    ) -> Result<(String, bool, u64)> {
        let plaintext_with_action = decrypt(key, ciphertext)?;
        static RE: Lazy<Regex> = Lazy::new(|| Regex::new(r"\((.+):(\d):(\d+)\)").unwrap());
        let (plaintext, action, validity) = match RE.captures(&plaintext_with_action) {
            Some(caps) => Ok((
                caps.get(1).map_or("", |m| m.as_str()).to_string(),
                caps.get(2).map_or("", |m| m.as_str()).parse::<i64>()? != 0,
                caps.get(3).map_or("", |m| m.as_str()).parse::<u64>()?,
            )),
            None => Err(anyhow!("No match found")),
        }?;
        Ok((plaintext, action, validity))
    }

    #[test]
    fn test_encrypt_decrypt_with_action_and_validity() {
        let plaintext = "hello".to_string();
        let action = true;
        let validity = false;
        let key = XChaCha20Poly1305::generate_key(&mut OsRng).to_vec();
        let ciphertext =
            encrypt_with_action_and_validity(&key, plaintext.clone(), action, validity).unwrap();
        let (decrypted, decrypted_action, decrypted_validity) =
            decrypt_with_action_and_validity(&key, ciphertext).unwrap();
        assert_eq!(plaintext, decrypted, "Plaintexts do not match");
        assert_eq!(action, decrypted_action, "Actions do not match");
        assert_eq!(0, decrypted_validity, "Validities do not match");
    }

    pub fn decrypt(key: &[u8], ciphertext: String) -> Result<String> {
        let cipher = Lazy::new(|| XChaCha20Poly1305::new_from_slice(key).unwrap());
        let cipher_bytes = BASE64_STANDARD.decode(ciphertext.clone())?;
        let nonce = XNonce::from_iter(cipher_bytes[0..24].iter().cloned());
        let remaining = &cipher_bytes[24..];
        let plaintext = match cipher.decrypt(&nonce, remaining) {
            Ok(p) => Ok(str::from_utf8(&p)?.to_string()),
            Err(e) => Err(anyhow!("Decrypt error {}", e.to_string())),
        }?;
        Ok(plaintext)
    }
}
pub mod hash {
    use anyhow::Result;

    use hmac::{Hmac, Mac};
    use sha2::Sha256;
    use tiger::{Digest, Tiger};
    type HmacSha256 = Hmac<Sha256>;

    pub fn prf(key: &[u8], plaintext: String) -> Result<String> {
        let mut mac = HmacSha256::new_from_slice(key)?;

        mac.update(&plaintext.into_bytes());
        Ok(hex::encode(mac.finalize().into_bytes()))
    }

    pub fn hash(plaintext: String) -> Result<String> {
        let mut hasher = Tiger::new();
        hasher.update(plaintext.as_bytes());
        Ok(hex::encode(hasher.finalize()))
    }

    pub fn hash_with_counter(plaintext: String, counter: usize) -> Result<String> {
        hash(format!("{}{}", plaintext, counter))
    }

    pub fn hash_with_counter_and_nonce(
        plaintext: String,
        counter: usize,
        nonce: &String,
    ) -> Result<String> {
        // let nonce =
        hash(format!("{}{}{}", nonce, plaintext, counter))
    }
}
pub mod keys {
    use anyhow::Result;
    use chacha20poly1305::{aead::OsRng, KeyInit, XChaCha20Poly1305};
    use log::debug;
    use std::{fs::File, io::Read, os::unix::fs::MetadataExt, path::Path};
    // function to load key from file
    pub fn load_from_file(filepath: &str) -> Result<Vec<u8>> {
        let fp = Path::new(filepath);
        // if file exists, load from file
        // if not, create key and write to file
        match fp.exists() {
            true => {
                debug!("{:?} exists, loading from file", fp);
                let mut file = File::open(fp)?;
                let metadata = file.metadata()?;
                let mut buffer = vec![0; metadata.size() as usize];
                file.read_exact(&mut buffer)?;
                Ok(buffer)
            }
            false => {
                debug!("{:?} does not exist, generating", fp);
                // generate key
                let key = XChaCha20Poly1305::generate_key(&mut OsRng).to_vec();
                // write to file
                std::fs::write(fp, &key)?;
                Ok(key)
            }
        }
    }
}
