use axum::{
    extract::{self, DefaultBodyLimit, Json, Path},
    http::StatusCode,
    routing::{get, post, put},
    Router,
};
use crypto_utils::{
    chacha::{
        rand::{seq::SliceRandom, SeedableRng},
        ChaCha20Rng,
    },
    hash::hash_with_counter_and_nonce,
};
mod cli;
use cli::Cli;
use serde::{Deserialize, Serialize};
use serde_json::{json, Value};
use std::{
    collections::HashMap,
    fs::{self, OpenOptions},
    io::Write,
};
use tower_http::trace::{self, TraceLayer};
use tracing::{info, Level};

#[derive(Serialize, Deserialize)]
struct ED {
    db: HashMap<String, String>,
}

#[derive(Serialize, Deserialize)]
struct Init {
    level: usize,
    db: HashMap<String, String>,
    nonce: String,
    metadata: String,
    btree: String,
}

#[derive(Clone)]
pub struct ServerConfig {
    pub port: u16,
    pub index: usize,
}

#[tokio::main]
async fn main() {
    let cli = <Cli as clap::Parser>::parse();

    let config = ServerConfig {
        port: cli.port,
        index: cli.index,
    };

    tracing_subscriber::fmt()
        .with_target(false)
        .compact()
        .init();

    let listener = tokio::net::TcpListener::bind(format!("127.0.0.1:{}", config.port))
        .await
        .unwrap();
    info!("listening on {}", listener.local_addr().unwrap());

    let app = Router::new()
        .route("/", post(init))
        .route("/key/{query}", get(search))
        .route("/key", put(insert))
        .route("/sda", put(sda))
        .route("/sda/{level}", get(get_sda))
        .route("/btrees", get(get_btrees))
        .route("/metadata", post(init_metadata))
        .route("/metadata", get(get_metadata))
        .layer(DefaultBodyLimit::disable())
        .layer(
            TraceLayer::new_for_http()
                .make_span_with(trace::DefaultMakeSpan::new().level(Level::INFO))
                .on_response(trace::DefaultOnResponse::new().level(Level::INFO)),
        )
        .with_state(config);
    axum::serve(listener, app).await.unwrap();
}

async fn init_metadata(
    extract::State(config): extract::State<ServerConfig>,
    payload: String,
) -> Result<(), StatusCode> {
    let file_name = format!("metadata_{}", config.index);
    match fs::write(file_name, payload) {
        Ok(_) => Ok(()),
        Err(x) => {
            println!("{:#?}", x);
            Err(StatusCode::INTERNAL_SERVER_ERROR)
        }
    }
}

/// Initializes the database from a json payload.
///
/// Dumps the contents into `sda_{level}.json`.
async fn init(
    extract::State(config): extract::State<ServerConfig>,
    extract::Json(payload): extract::Json<Init>,
) -> Result<(), StatusCode> {
    let file_name = format!("sda_{}_{}.json", config.index, payload.level);
    match fs::write(
        file_name,
        serde_json::from_value::<Value>(json!(payload.db))
            .unwrap()
            .to_string(),
    ) {
        Ok(_) => Ok(()),
        Err(x) => {
            println!("{:#?}", x);
            Err(StatusCode::INTERNAL_SERVER_ERROR)
        }
    }?;
    let file_name = format!("metadata_{}", config.index);
    match fs::write(file_name, payload.metadata) {
        Ok(_) => Ok(()),
        Err(x) => {
            println!("{:#?}", x);
            Err(StatusCode::INTERNAL_SERVER_ERROR)
        }
    }?;
    let file_name = format!("btree_{}_{}", config.index, payload.level);
    match fs::write(file_name, payload.btree) {
        Ok(_) => Ok(()),
        Err(x) => {
            println!("{:#?}", x);
            Err(StatusCode::INTERNAL_SERVER_ERROR)
        }
    }?;
    let nonce_name = format!("nonce_{}_{}", config.index, payload.level);
    match fs::write(nonce_name, payload.nonce) {
        Ok(_) => Ok(()),
        Err(_) => Err(StatusCode::INTERNAL_SERVER_ERROR),
    }
}
async fn get_btrees(
    extract::State(config): extract::State<ServerConfig>,
) -> Result<Json<Value>, StatusCode> {
    let mut ret_vec: Vec<String> = Vec::new();
    for i in 0..64 {
        let file_name = format!("btree_{}_{}", config.index, i);
        // check if file exists
        if !std::path::Path::new(&file_name).exists() {
            continue;
        }
        let data = match fs::read_to_string(file_name) {
            Ok(x) => x,
            Err(x) => {
                info!("{:#?} on loop {}", x, i);
                continue;
            }
        };
        ret_vec.push(data);
    }
    // shuffle vector
    ret_vec.shuffle(&mut crypto_utils::chacha::ChaCha20Rng::from_os_rng());
    Ok(Json(json!(ret_vec)))
}

#[derive(Serialize, Deserialize)]
struct Sda {
    level: usize,
    db: HashMap<String, String>,
    nonce: String,
    btree: String,
}

async fn get_sda(
    extract::Path(level): extract::Path<usize>,
    extract::State(config): extract::State<ServerConfig>,
) -> Result<Json<bool>, StatusCode> {
    let file_name = format!("sda_{}_{}.json", config.index, level);
    let r = !std::path::Path::new(&file_name).exists();
    Ok(Json(r))
}
async fn sda(
    extract::State(config): extract::State<ServerConfig>,
    extract::Json(payload): extract::Json<Sda>,
) -> Result<Json<InsertResults>, StatusCode> {
    let sda_path = format!("sda_{}_{}.json", config.index, payload.level);
    let btree_path = format!("btree_{}_{}", config.index, payload.level);
    let nonce_path = format!("nonce_{}_{}", config.index, payload.level);
    let sda_exists = std::path::Path::new(&sda_path).exists();
    // if that bin doesn't already exist, we create a it
    if !sda_exists {
        // write payload
        let mut writer = OpenOptions::new()
            .write(true)
            .create(true)
            .truncate(true)
            .open(&sda_path)
            .unwrap();
        if let Err(x) = writer.write_all(json!(payload.db).to_string().as_bytes()) {
            info!("{:#?}", x);
            return Err(StatusCode::INTERNAL_SERVER_ERROR);
        };
        // write btree
        let mut writer = OpenOptions::new()
            .write(true)
            .create(true)
            .truncate(true)
            .open(btree_path)
            .unwrap();
        if let Err(x) = writer.write_all(payload.btree.as_bytes()) {
            info!("{:#?}", x);
            return Err(StatusCode::INTERNAL_SERVER_ERROR);
        };
        let mut writer = OpenOptions::new()
            .write(true)
            .create(true)
            .truncate(true)
            .open(nonce_path)
            .unwrap();
        if let Err(x) = writer.write_all(payload.nonce.as_bytes()) {
            info!("{:#?}", x);
            return Err(StatusCode::INTERNAL_SERVER_ERROR);
        };
        return Ok(Json(InsertResults {
            sda: false,
            values: vec![],
            sda_level: payload.level,
        }));
    }
    // info!("sda_{} exists", payload.level);
    // we cluster the bins
    let data_string = match fs::read_to_string(&sda_path) {
        Ok(x) => x,
        Err(x) => {
            info!("{:#?}", x);
            return Err(StatusCode::INTERNAL_SERVER_ERROR);
        }
    };
    let mut sda_existing_value = serde_json::from_str::<HashMap<String, String>>(&data_string)
        .unwrap()
        .values()
        .map(|x| x.to_owned())
        .collect::<Vec<String>>();
    // extend with payload
    sda_existing_value.extend(payload.db.values().map(|x| x.to_owned()));
    sda_existing_value.shuffle(&mut ChaCha20Rng::from_os_rng());
    // remove existing files
    fs::remove_file(&sda_path).unwrap();
    fs::remove_file(&btree_path).unwrap();
    fs::remove_file(&nonce_path).unwrap();
    let response = Json(InsertResults {
        sda: true,
        values: sda_existing_value,
        sda_level: payload.level + 1,
    });
    // info!(
    //     "Requesting new sda with requeset {:?} and level {}",
    //     response,
    //     payload.level + 1
    // );
    Ok(response)
}
async fn get_metadata(
    extract::State(config): extract::State<ServerConfig>,
) -> Result<String, StatusCode> {
    let data = match fs::read_to_string(format!("metadata_{}", config.index)) {
        Ok(x) => x,
        Err(x) => {
            info!("{:#?}", x);
            return Err(StatusCode::INTERNAL_SERVER_ERROR);
        }
    };
    // info!("Returning metadata {}", data);
    Ok(data)
}
#[derive(Serialize, Deserialize, Debug)]
struct InsertResults {
    sda: bool,
    values: Vec<String>,
    sda_level: usize,
}

#[derive(Serialize, Deserialize)]
struct InsertRequest {
    db: HashMap<String, String>,
    btree: String,
    nonce: String,
}
async fn insert(
    extract::State(config): extract::State<ServerConfig>,
    extract::Json(payload): extract::Json<InsertRequest>,
) -> Result<Json<InsertResults>, StatusCode> {
    let sda0_path = format!("sda_{}_0.json", config.index);
    let btree0_path = format!("btree_{}_0", config.index);
    let nonce0_path = format!("nonce_{}_0", config.index);
    let sda_0_exists = std::path::Path::new(&sda0_path).exists();
    // need to update sda
    if sda_0_exists {
        let data_string = fs::read_to_string(&sda0_path).unwrap_or_default();
        let mut sda_0_existing_value =
            serde_json::from_str::<HashMap<String, String>>(&data_string)
                .unwrap()
                .values()
                .map(|x| x.to_owned())
                .collect::<Vec<String>>();
        // payload has only one value
        sda_0_existing_value.push(payload.db.values().next().unwrap().to_owned());
        sda_0_existing_value.shuffle(&mut ChaCha20Rng::from_os_rng());
        let respond_json = Json(InsertResults {
            sda: true,
            values: sda_0_existing_value,
            sda_level: 1,
        });
        //delete file
        fs::remove_file(sda0_path).unwrap();
        fs::remove_file(btree0_path).unwrap();
        fs::remove_file(nonce0_path).unwrap();
        return Ok(respond_json);
    }
    // write database
    let mut writer = OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(&sda0_path)
        .unwrap();
    writer
        .write_all(json!(payload.db).to_string().as_bytes())
        .unwrap();
    // write btree
    let mut writer = OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(btree0_path)
        .unwrap();
    writer.write_all(payload.btree.as_bytes()).unwrap();
    let mut writer = OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(nonce0_path)
        .unwrap();
    writer.write_all(payload.nonce.as_bytes()).unwrap();
    Ok(Json(InsertResults {
        sda: false,
        values: vec![],
        sda_level: 0,
    }))
}

/// Searches for database entries matching the `key` path parameter.
///
/// Hashes the key with a counter until no more results are found.
///
/// Returns json in the form:
///
/// ```json
/// {
/// ctr: "ciphertext"
/// }
/// ```
async fn search(
    Path(key): Path<String>,
    extract::State(config): extract::State<ServerConfig>,
) -> Result<Json<Value>, StatusCode> {
    // search now looks thorugh all sda bins
    // we note that a given key is valid in possibly multiple bins,
    let mut ret_vec: Vec<String> = Vec::new();
    for i in 0..64 {
        let file_name = format!("sda_{}_{}.json", config.index, i);
        // check if file exists
        if !std::path::Path::new(&file_name).exists() {
            continue;
        }
        let data = match fs::read_to_string(file_name) {
            Ok(x) => x,
            Err(x) => {
                info!("{:#?} on loop {}", x, i);
                continue;
            }
        };
        let file_name = format!("nonce_{}_{}", config.index, i);
        let nonce = match fs::read_to_string(file_name) {
            Ok(x) => x,
            Err(x) => {
                info!("{:#?} on loop {}", x, i);
                continue;
            }
        };
        let db: HashMap<String, String> = serde_json::from_str(&data).unwrap();
        let mut n = 0;
        while let Some(ciphertext) =
            db.get(&hash_with_counter_and_nonce(key.clone(), n, &nonce).unwrap())
        {
            info!("Found {}({}) in sda level {} ", key, n, i);
            ret_vec.push(ciphertext.to_owned());
            n += 1;
        }
    }
    if !ret_vec.is_empty() {
        return Ok(Json(json!(ret_vec)));
    }
    Err(StatusCode::NOT_FOUND)
}
