use axum::{
    extract::{DefaultBodyLimit, Json, Path},
    http::StatusCode,
    routing::{get, post},
    Router,
};
use crypto_utils::hash::hash_with_counter;
use serde::{Deserialize, Serialize};
use serde_json::{json, Value};
use std::{collections::HashMap, fs};
use tower_http::trace::{self, TraceLayer};
use tracing::Level;

#[derive(Serialize, Deserialize)]
struct ED {
    db: HashMap<String, String>,
}

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt()
        .with_target(false)
        .compact()
        .init();

    let app = Router::new()
        .route("/", post(init))
        .layer(DefaultBodyLimit::disable())
        .route("/key/{query}", get(search))
        .layer(
            TraceLayer::new_for_http()
                .make_span_with(trace::DefaultMakeSpan::new().level(Level::INFO))
                .on_response(trace::DefaultOnResponse::new().level(Level::INFO)),
        );

    let listener = tokio::net::TcpListener::bind("127.0.0.1:8081")
        .await
        .unwrap();
    axum::serve(listener, app).await.unwrap();
}

/// Initializes the database from a json payload.
///
/// Dumps the contents into `db.json`.
async fn init(Json(payload): Json<Value>) -> Result<(), StatusCode> {
    match fs::write("./db.json", payload.to_string()) {
        Ok(_) => Ok(()),
        Err(x) => {
            println!("{:#?}", x);
            Err(StatusCode::INTERNAL_SERVER_ERROR)
        }
    }
}

/// Searches for database entries matching the `key` path parameter.
///
/// Hashes the key with a counter until no more results are found.
///
/// Returns json in the form:
///
/// ```json
/// {
/// ctr: "ciphertext"
/// }
/// ```
async fn search(Path(key): Path<String>) -> Result<Json<Value>, StatusCode> {
    // Read the database from db.json
    let data_string = fs::read_to_string("./db.json").unwrap_or_default();
    // Parse the database
    let ed: ED = ED {
        db: {
            if let Ok(db) = serde_json::from_str(&data_string) {
                db
            } else {
                return Err(StatusCode::INTERNAL_SERVER_ERROR);
            }
        },
    };
    let mut n = 0;
    let mut ids: HashMap<usize, &String> = HashMap::new();
    // The actual search
    while let Some(ciphertext) = ed.db.get(&hash_with_counter(key.clone(), n).unwrap()) {
        ids.insert(n, ciphertext);
        n += 1;
    }
    Ok(Json(json!(ids)))
}
